# -*- coding: utf-8 -*-
import pandas as pd
from zipline.api import order_target_percent, record, symbol
import pytz
import zipline
from datetime import datetime
from contrib_lib.quant.utils.chinese_stock_calendar import ChineseStockCalendar
# Use a random forest classifier. More here: http://scikit-learn.org/stable/user_guide.html
from sklearn.ensemble import RandomForestClassifier
from collections import deque
import numpy as np
from collections import OrderedDict
from contrib_lib.quant.zipline.default_strategy import DefaultStrategy


class SimpleMachineLearningStrategy(DefaultStrategy):
    """
    机器学习算法交易
    参考：https://www.quantopian.com/posts/simple-machine-learning-example
    进化策略：https://www.quantopian.com/posts/evolutionary-strategy
    """
    def __init__(self):
        pass

    def prepare_data(self):
        data = OrderedDict()
        data['SPY'] = pd.read_csv('/Users/tony/zipline_data_can_delete/AAPL.csv', index_col=0,
                                  parse_dates=[['Date', 'Timestamp']])
        data['SPY'] = data['SPY'][['OpenPrice', 'HighPrice', 'LowPrice', 'ClosePrice', 'TotalVolume']]
        data['SPY'] = data['SPY'].drop(columns=['HighPrice', 'LowPrice', 'ClosePrice', 'TotalVolume'])
        # data['SPY'] = data['SPY'].loc(axis=0)[:, "OpenPrice"]  # TODO 只选取price这1列
        panel = pd.Panel(data)
        panel.major_axis = panel.major_axis.tz_localize(pytz.utc)
        # panel.minor_axis = ['open', 'high', 'low', 'close', 'volume']
        panel.minor_axis = ['open']
        return panel

    def initialize(self, context):
        context.security = symbol("SPY")
        context.window_length = 10  # Amount of prior bars to study
        context.classifier = RandomForestClassifier()  # Use a random forest classifier

        # deques are lists with a maximum length where old entries are shifted out
        context.recent_prices = deque(maxlen=context.window_length + 2)  # Stores recent prices
        context.X = deque(maxlen=500)  # Independent, or input variables
        context.Y = deque(maxlen=500)  # Dependent, or output variable

        context.prediction = 0  # Stores most recent prediction

    def handle_data(self, context, data):
        context.recent_prices.append(data[context.security].price)  # Update the recent prices
        if len(context.recent_prices) == context.window_length + 2:  # If there's enough recent price data
            # Make a list of 1's and 0's, 1 when the price increased from the prior bar
            changes = np.diff(context.recent_prices) > 0
            context.X.append(changes[:-1])  # Add independent variables, the prior changes
            context.Y.append(changes[-1])  # Add dependent variable, the final change
            if len(context.Y) >= 100:  # There needs to be enough data points to make a good model
                context.classifier.fit(context.X, context.Y)  # Generate the model
                context.prediction = context.classifier.predict(changes[1:])  # Predict
                # If prediction = 1, buy all shares affordable, if 0 sell all shares
                order_target_percent(context.security, context.prediction)
                record(prediction=int(context.prediction))

    def run_algorithm(self):
        start_time = datetime(2018, 1, 2, 9, 31, 0, 0, pytz.utc)
        end_time = datetime(2018, 2, 4, 16, 0, 0, 0, pytz.utc)
        zipline.run_algorithm(start=start_time,
                                     end=end_time,
                                     initialize=self.initialize,
                                     capital_base=100000,
                                     handle_data=self.handle_data,
                                     before_trading_start=self.before_trading_start,
                                     data_frequency='minute',
                                     data=self.prepare_data(),
                                     trading_calendar=ChineseStockCalendar(),
                                     analyze=self.analyze)


if __name__ == '__main__':
    SimpleMachineLearningStrategy().run_algorithm()




# -*- coding: utf-8 -*-
from contrib.quant.zipline.default_strategy import DefaultStrategy
from contrib.quant.utils.chinese_stock_calendar import ChineseStockCalendar
import datetime


class DataCampStrategy(DefaultStrategy):
    """
    Python for Finance Tutorial For Beginners
    https://github.com/datacamp/datacamp-community-tutorials/blob/master/Python%20Finance%20Tutorial%20For%20Beginners/Python%20For%20Finance%20Beginners%20Tutorial.ipynb
    """
    def __init__(self):
        super().__init__()

    def prepare_data(self):
        from collections import OrderedDict
        data = OrderedDict()
        from pandas_datareader import data as pdr
        data['SPY'] = pdr.get_data_yahoo('AAPL',
                                  start=datetime.datetime(2018, 1, 2),
                                  end=datetime.datetime(2018, 5, 10))
        data['SPY'].to_csv('~/.zipline/data/treasury_curves.csv')  # 手工生成treasury_curves.csv文件
        data['SPY'] = data['SPY'][['Open', 'High', 'Low', 'Close', 'Volume']]
        panel = self.pandas.Panel(data)
        panel.major_axis = panel.major_axis.tz_localize(tz='utc')
        panel.minor_axis = ['open', 'high', 'low', 'close', 'volume']
        return panel

    def initialize(self, context):
        super().initialize(context)
        context.sym = self.zipline.api.symbol('SPY')
        print("22222", context.sym)

    def before_trading_start(self, context, data):
        super().before_trading_start(context, data)

    def handle_data(self, context, data):
        super().handle_data(context, data)

    def analyze(self, context, records):
        super().analyze(context, records)

    def run_algorithm(self):
        start_time = self.pandas.Timestamp('2018-01-02 00:00:00', tz='utc')
        end_time = self.pandas.Timestamp('2018-05-10 00:00:00', tz='utc')
        data_frequency = 'daily'
        perf = self.zipline.run_algorithm(start=start_time,
                              end=end_time,
                              initialize=self.initialize,
                              capital_base=100000,
                              handle_data=self.handle_data,
                              before_trading_start=self.before_trading_start,
                              data_frequency=data_frequency,
                              data=self.prepare_data(),
                              trading_calendar=ChineseStockCalendar(data_frequency=data_frequency),
                              analyze=self.analyze)
        return perf


if __name__ == '__main__':
    DataCampStrategy().run_algorithm()




# -*- coding: utf-8 -*-
from contrib_lib.quant.utils.chinese_stock_calendar import ChineseStockCalendar
from contrib_lib.quant.zipline.default_strategy import DefaultStrategy
from py_common_util.common.common_utils import CommonUtils
import matplotlib.pyplot as plt
from matplotlib.dates import DateFormatter
from empyrical import cum_returns, annual_return, sharpe_ratio, max_drawdown, alpha, beta
import pandas as pd


class DualMovingAverageStrategy(DefaultStrategy):
    """
    参考：https://github.com/quantopian/zipline/blob/master/zipline/examples/dual_moving_average.py
    """
    def __init__(self):
        super().__init__()

    def prepare_data(self):
        super().prepare_data()
        from collections import OrderedDict
        data = OrderedDict()
        csv_file_local = "/Users/tony/zipline_data_can_delete/AAPL.csv"
        csv_file_remote = "/tony/zipline_local_dataset/AAPL.csv"
        to_csv_file_local = "~/.zipline/data/treasury_curves.csv"
        to_csv_file_remote = "/tony/zipline_local_dataset/treasury_curves.csv"
        data['SPY'] = self.pandas.read_csv(csv_file_local, index_col=0, parse_dates=[['Date', 'Timestamp']])
        data['SPY'].to_csv(to_csv_file_local)  # 手工生成treasury_curves.csv文件
        data['SPY'] = data['SPY'][['OpenPrice', 'HighPrice', 'LowPrice', 'ClosePrice', 'TotalVolume']]
        panel = self.pandas.Panel(data)
        panel.major_axis = panel.major_axis.tz_localize(tz='utc')
        panel.minor_axis = ['open', 'high', 'low', 'close', 'volume']
        return panel

    def initialize(self, context):
        self.log.info("sid(int)=" + str(self.zipline.api.sid(0)))  # 这里的sid(0)等价symbol('SPY')
        context.sym = self.zipline.api.symbol('SPY')
        context.i = 0
        # Explicitly set the commission/slippage to the "old" value until we can
        # rebuild example data.
        # github.com/quantopian/zipline/blob/master/tests/resources/
        # rebuild_example_data#L105
        context.set_commission(self.zipline.finance.commission.PerShare(cost=.0075, min_trade_cost=1.0))
        context.set_slippage(self.zipline.finance.slippage.VolumeShareSlippage())
        # self.zipline.api.set_benchmark(self.zipline.api.symbol('SPY'))

    def before_trading_start(self, context, data):
        # 每个bar_open之前执行，对日K可选定当天待交易股票，分钟K可以用于初始化数据
        # self.log.info("before_trading_start...")
        pass

    def handle_data(self, context, data):
        # Skip first 300 days to get full windows
        context.i += 1
        if context.i < 300:
            return
        # Compute averages
        # history() has to be called with the same params
        # from above and returns a pandas dataframe.
        short_mavg = data.history(context.sym, 'price', 100, '1m').mean()
        long_mavg = data.history(context.sym, 'price', 300, '1m').mean()

        # Trading logic
        if short_mavg > long_mavg:
            # order_target orders as many shares as needed to
            # achieve the desired number of shares.
            self.zipline.api.order_target(context.sym, 100)
        elif short_mavg < long_mavg:
            self.zipline.api.order_target(context.sym, 0)

        # Save values for later inspection
        self.zipline.api.record(SPY=data.current(context.sym, "price"),
               short_mavg=short_mavg,
               long_mavg=long_mavg)

    def analyze(self, context, records):
        fig = plt.figure()
        ax1 = fig.add_subplot(211)
        records.portfolio_value.plot(ax=ax1)
        ax1.set_ylabel('Portfolio value (USD)')

        ax2 = fig.add_subplot(212)
        ax2.set_ylabel('Price (USD)')

        # If data has been record()ed, then plot it.
        # Otherwise, log the fact that no data has been recorded.
        if ('SPY' in records and 'short_mavg' in records and
                'long_mavg' in records):
            records['SPY'].plot(ax=ax2)
            records[['short_mavg', 'long_mavg']].plot(ax=ax2)

            trans = records.ix[[t != [] for t in records.transactions]]
            buys = trans.ix[[t[0]['amount'] > 0 for t in
                             trans.transactions]]
            sells = trans.ix[
                [t[0]['amount'] < 0 for t in trans.transactions]]
            ax2.plot(buys.index, records.short_mavg.ix[buys.index],
                     '^', markersize=10, color='m')
            ax2.plot(sells.index, records.short_mavg.ix[sells.index],
                     'v', markersize=10, color='k')
            plt.legend(loc=0)
        else:
            msg = 'SPY, short_mavg & long_mavg data not captured using record().'
            ax2.annotate(msg, xy=(0.1, 0.5))
            self.log.info(msg)
        plt.show()

    @CommonUtils.print_exec_time
    def run_algorithm(self):
        data = self.ts.proapi.stock_basic(exchange='', list_status='L', fields='ts_code,symbol,name,area,industry,list_date')
        self.log.info(data)
        start_time = self.pandas.Timestamp('2018-01-02 09:31:00', tz='utc')
        end_time = self.pandas.Timestamp('2018-02-04 16:00:00', tz='utc')
        data_frequency = 'minute'
        perf = self.zipline.run_algorithm(start=start_time,
                             end=end_time,
                             initialize=self.initialize,
                             capital_base=100000,
                             handle_data=self.handle_data,
                             before_trading_start=self.before_trading_start,
                             data_frequency=data_frequency,
                             data=self.prepare_data(),
                             trading_calendar=ChineseStockCalendar(data_frequency=data_frequency),
                             analyze=self.analyze)
        return perf


def draw_return_rate_line(result):
    import seaborn as sns
    sns.set_style('darkgrid')
    sns.set_context('notebook')
    ax = plt.axes()
    # 设置时间显示格式
    years_fmt = DateFormatter('%Y-%m-%d')
    ax.xaxis.set_major_formatter(years_fmt)
    # 让x轴坐标旋转45度
    labels = ax.get_xticklabels()
    plt.setp(labels, rotation=35, horizontalalignment='right')
    # 画出收益率曲线
    sns.lineplot(x='period_close',
    y='algorithm_period_return',
    data=result,
    label="AAPL")
    sns.lineplot(x='period_close',
        y='benchmark_period_return',
        data=result, label="SPX")

    plt.legend(loc='upper left')
    plt.title("return rate of AAPL and SPX")
    plt.xlabel('time')
    plt.ylabel('return rate')
    plt.show()


if __name__ == '__main__':
    # zipline风险指标计算 (empyrical模块)  https://www.cnblogs.com/fangbei/p/8432891.html
    # https://github.com/zhanghan1990/zipline-chinese/blob/master/zipline/examples/stock_select.py
    perf = DualMovingAverageStrategy().run_algorithm()
    # 画出收益曲线图
    draw_return_rate_line(perf)
    return_list = perf['returns']
    # 计算年化收益率
    ann_return = annual_return(return_list)
    # 计算累计收益率
    cum_return_list = cum_returns(return_list)
    # 计算sharp ratio
    sharp = sharpe_ratio(return_list)
    # 最大回撤
    max_drawdown_ratio = max_drawdown(return_list)
    print("年化收益率 = {:.2%}, 累计收益率 = {:.2%}, 最大回撤 = {:.2%}, 夏普比率 = {:.2f} ".format
          (ann_return, cum_return_list[-1], max_drawdown_ratio, sharp))

    returns = pd.Series(
        index=pd.date_range('2017-03-10', '2017-03-19'),
        data=(-0.012143, 0.045350, 0.030957, 0.004902, 0.002341, -0.02103, 0.00148, 0.004820, -0.00023, 0.01201)
    )
    benchmark_returns = pd.Series(
        index=pd.date_range('2017-03-10', '2017-03-19'),
        data=(-0.031940, 0.025350, -0.020957, -0.000902, 0.007341, -0.01103, 0.00248, 0.008820, -0.00123, 0.01091)
    )
    alpha_return = alpha(returns=returns, factor_returns=benchmark_returns, risk_free=0.01)
    beta_return = beta(returns=returns, factor_returns=benchmark_returns, risk_free=0.01)
    print("alpha_return", alpha_return)
    print("\nbeta_return", beta_return)







from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import logging
import time
import ray
from ray.experimental.streaming.communication import QueueConfig
from ray.experimental.streaming.streaming import Environment
from ray.experimental.streaming.batched_queue import BatchedQueue
from ray.experimental.streaming.operator import OpType, PStrategy
from py_common_util.common.date_utils import DateUtils

# define logger
logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


def splitter(line):
    # return line.split()
    return line + "-by-splitter"


def filter_fn(word):
    if "f" in word:
        return True
    return False


class CustormConfig(object):
    def __init__(self, parallelism=1):
        self.queue_config = QueueConfig(
            max_size=999999,
            max_batch_size=99999,
            max_batch_time=0.01,
            prefetch_depth=10,
            background_flush=False)
        self.parallelism = parallelism


if __name__ == "__main__":
    ray.init(redis_address="192.168.95.27:6379")
    ray.register_custom_serializer(BatchedQueue, use_pickle=True)
    ray.register_custom_serializer(OpType, use_pickle=True)
    ray.register_custom_serializer(PStrategy, use_pickle=True)
    env = Environment(CustormConfig())

    def attribute_selector(tuple):
        return tuple[1]

    def key_selector(tuple):
        return tuple[0]

    class CustormSource:
        def __init__(self):
            self._count = 0

        def get_next(self):
            self._count += 1
            if self._count % 20 == 0:
                # print("***get_next_nvl***:" + str(self._count))
                return str(self._count) + "-uuunited\nStates-" + DateUtils.now_to_str()
            else:
                return str(self._count) + "-United States-" + DateUtils.now_to_str()

    def print_log(content):
        logger.info(content)
        time.sleep(10)
        logger.info(content + ",finished:" + DateUtils.now_to_str())

    stream = env.source(CustormSource()) \
                .set_parallelism(10) \
                .inspect(print_log)\
                .set_parallelism(10)

    # stream = env.read_text_file(args.input_file) \
    #             .shuffle() \
    #             .flat_map(splitter) \
    #             .set_parallelism(2) \
    #             .key_by(key_selector) \
    #             .sum(attribute_selector) \
    #             .inspect(print)  # Prints the contents of the
    start = time.time()
    env_handle = env.execute()
    ray.get(env_handle)
    end = time.time()
    logger.info("Elapsed time: {} secs".format(end - start))
    logger.info("Output stream id: {}".format(stream.id))
    print("end!")



# -*- coding: utf-8 -*-
from abc import abstractmethod, ABCMeta
from six import with_metaclass
# import quandl
import zipline as zl
import pandas as pd
import numpy as np
import sys
from logbook import Logger, StreamHandler, FileHandler
# from skydl.zipline.tushare_dataset import TushareDataset


class AbstractStrategy(with_metaclass(ABCMeta)):
    """
    zipline数据默认缓存在：~/.zipline/data/
    10 Minutes to pandas  https://pandas.pydata.org/pandas-docs/stable/getting_started/10min.html
    https://github.com/quantopian/zipline
    量化交易实践篇（1）——量化交易平台Quantopian介绍 https://www.jianshu.com/p/5925f4f13a84
    量化交易实践篇（2）—— Quantopian策略实现初体验  https://www.jianshu.com/p/5f7a8f53045d
    ===============
    fix issue: Can't run zipline backtest without treasury data
    https://github.com/quantopian/zipline/issues/2422
    https://github.com/nateGeorge/treasury_data_backup
    ...
    data['SPY'].to_csv('~/.zipline/data/treasury_curves.csv')  # 手工生成treasury_curves.csv文件
    ...
    ===============
    """
    @property
    def log(self):
        return self._log

    @log.setter
    def log(self, value):
        self._log = value

    # @property
    # def ts(self):
    #     return self._ts

    @property
    def zipline(self):
        return self._zipline

    @property
    def pandas(self):
        return self._pd

    @property
    def numpy(self):
        return self._np

    def __init__(self):
        """
        Python日志记录包logbook https://blog.yasking.org/a/python-logbook.html
        记录日志到文件和STDOUT
        使用StreamHandler记录的日志会以流输出，这里指定sys.stdout也就是记录到标准输出，与print一样
        StreamHandler(sys.stdout, level='DEBUG').push_application()
        FileHandler('app.log', bubble=True, level='INFO').push_application()
        """
        StreamHandler(sys.stdout).push_application()
        self._log = Logger(self.__class__.__name__)
        # quandl.ApiConfig.api_key = ""  # 不可向外泄露api_key
        # self.log.info("quandl version: " + quandl.version.VERSION)
        # self._ts = TushareDataset(log=self._log)
        self._zipline = zl
        self.log.info("zipline version: " + self.zipline.__version__)
        self._pd = pd
        self.log.info("pandas version: " + self.pandas.__version__)
        self._np = np
        self.log.info("numpy version: " + self.numpy.__version__)

    @abstractmethod
    def prepare_data(self):
        # zipline数据默认缓存在：~/.zipline/data/
        return None

    @abstractmethod
    def initialize(self, context):
        # 启动后需要处理的一次性逻辑
        self.log.info("initialize...")

    @abstractmethod
    def before_trading_start(self, context, data):
        # 每个bar_open之前执行，对日K可选定当天待交易股票，分钟K可以用于初始化数据
        # self.log.info("before_trading_start...")
        pass

    @abstractmethod
    def handle_data(self, context, data):
        # 定时执行，处理当前周期中待处理订单
        # self.log.info("handle_data...")
        pass

    @abstractmethod
    def analyze(self, context, records):
        self.log.info("analyze...")

    @abstractmethod
    def run_algorithm(self):
        """
        调用例子：
        start_time = self.pandas.Timestamp('2018-01-02 09:31:00', tz='utc')
        end_time = self.pandas.Timestamp('2018-02-04 16:00:00', tz='utc')
        data_frequency = 'minute'
        perf = zipline.run_algorithm(start=start_time,
                             end=end_time,
                             initialize=self.initialize,
                             capital_base=100000,
                             handle_data=self.handle_data,
                             before_trading_start=self.before_trading_start,
                             data_frequency=data_frequency,
                             data=self.prepare_data(),
                             trading_calendar=ChineseStockCalendar(data_frequency=data_frequency),
                             analyze=self.analyze)
        return perf
        """
        self.log.info("run_algorithm...")
        return None


if __name__ == '__main__':
    AbstractStrategy().run_algorithm()



# -*- coding: utf-8 -*-
import tushare as ts
from logbook import Logger, StreamHandler, FileHandler
import sys


class TushareDataset:
    """
    https://tushare.pro/
    文档说明https://tushare.pro/document/1
    token来自 https://tushare.pro/user/token
    """
    @property
    def log(self):
        return self._log

    @property
    def proapi(self):
        return self._proapi

    def __init__(self, log=None, ts_api_token=""):
        if log:
            self._log = log
        else:
            StreamHandler(sys.stdout).push_application()
            self._log = Logger(self.__class__.__name__)
        self.log.info("tushare version: " + ts.__version__)
        self._proapi = ts.pro_api(ts_api_token)


if __name__ == '__main__':
    ts = TushareDataset()
    data = ts.proapi.stock_basic(exchange='', list_status='L', fields='ts_code,symbol,name,area,industry,list_date')
    ts.log.info(data)
    df = ts.proapi.trade_cal(exchange='', start_date='20190201', end_date='20190303',
                               fields='exchange,cal_date,is_open,pretrade_date', is_open='0')
    ts.log.info(df)



# -*- coding: utf-8 -*-
from datetime import time, datetime
from trading_calendars import TradingCalendar
from zipline.utils.memoize import lazyval
from pandas.tseries.offsets import CustomBusinessDay
import pytz
from pytz import timezone


class ChineseStockCalendar(TradingCalendar):
    """
    for creating and registering our calendar
    Exchange calendar for 24/7 trading.
    Open Time: 12am, UTC
    Close Time: 11:59pm, UTC
    参考：https://github.com/zhanghan1990/zipline-chinese/blob/master/zipline/utils/tradingcalendar_china.py
    """
    @property
    def data_frequency(self):
        return self._data_frequency

    def __init__(self, data_frequency='minute'):
        self._data_frequency = data_frequency  # 'minute' or 'daily'
        super().__init__()

    @property
    def name(self):
        return "chinese_stock_calendar"

    @property
    def tz(self):
        return timezone("UTC")

    @property
    def open_times(self):
        return [(None, time(9, 31))] if self.data_frequency == 'minute' else [(None, time(0))]

    @property
    def close_times(self):
        return [(None, time(16))] if self.data_frequency == 'minute' else [(None, time(0))]

    @lazyval
    def day(self):
        return CustomBusinessDay(
            holidays=self.adhoc_holidays,
            weekmask='Mon Tue Wed Thu Fri',  # Mon Tue Wed Thu Fri Sat Sun
        )

    @property
    def adhoc_holidays(self):
        return [datetime(2018, 1, 15, tzinfo=pytz.utc)]



# -*- coding: utf-8 -*-
import ray
import logging
import traceback
import time
from py_common_util.common.date_utils import DateUtils
from ray.experimental.streaming.streaming import Environment, DataStream
from skydl.ray.experimental.ray_streaming_util import DefaultStreamingSource, DefaultStreamingEnvConfig

# define logger
logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


class RayStreamingExecutor:
    """
    ray流计算执行器，子类可以重写build_stream()方法
    """
    def __init__(self):
        logger.info("starting a streaming executor...")

    def execute(self, streaming_parallelism=1, env=None):
        try:
            if not env:
                env = Environment(DefaultStreamingEnvConfig(parallelism=streaming_parallelism))
            stream = self.build_stream(streaming_parallelism=streaming_parallelism, env=env)
            self.execute_forever(env, stream)
        except Exception as e:
            logger.error("RayStreamingExecutor fail on execute(): %s", str(e))
            traceback.print_exc()

    def execute_forever(self, env, stream, loop_idle_seconds=1):
        while(True):
            try:
                ray.get(env.execute())
                logger.info("Ray Output stream id: {}".format(stream.id))
                logger.info("a round of ray streaming is finished!!!")
            except Exception as e:
                logger.error("Error Occurred on Ray Streaming: %s", str(e))
                traceback.print_exc()
                # TODO 发送运维警报
                time.sleep(30)  # 30秒后重启streaming
            finally:
                time.sleep(loop_idle_seconds)  # 等待所有算子的数据都执行完，然后休息1秒后再继续重新执行同一个stream计算图

    def build_stream(self, streaming_parallelism, env)->DataStream:
        """
        子类需要重写该方法构建具体的业务stream
        注意streaming算子类必须要预先install在所有节点中，或者都在该方法里定义
        @:param streaming_parallelism default value is 1
        @:param env
        @:return DataStream
        """
        class FooStreamingSource(DefaultStreamingSource):
            def __init__(self):
                logger.info("FooStreamingSource init...(init func only be executed 1 time in a stream graph!!!)")
                self._count = 0

            def get_next(self):
                self._count += 1
                logger.info("get_next:" + str(self._count))
                if self._count % 2 == 0:
                    time.sleep(5)  # with env config background_flush=True, even if it sleep ever, it also success to auto flush。but can not kill current stream until return None。
                    # return None
                    return str(self._count) + "-aaa\nbbb*"+DateUtils.now_to_str()
                else:
                    return str(self._count) + "-ccc*"+DateUtils.now_to_str()

        def splitter_fn(record):
            """convert record to another record list, e.g. [record]"""
            logger.info("splitter: " + record)
            # return [record+"-splitter-" + str(i) for i in range(20)]
            return [record]

        def handle_record_fn(record):
            """处理回测的逻辑"""
            logger.info("record=" + record)
            time.sleep(10)
            logger.info(record + ", finished:" + DateUtils.now_to_str())

        stream = env.source(
            FooStreamingSource()
        ).set_parallelism(2).flat_map(
            flatmap_fn=splitter_fn
        ).set_parallelism(1).map(
            map_fn=handle_record_fn,
            name="nocode_backtest_map"
        ).set_parallelism(2)
        return stream













# -*- coding: utf-8 -*-
from ray.experimental.streaming.communication import QueueConfig


class DefaultStreamingEnvConfig(object):
    """
    default ray streaming env config
    """
    def __init__(self, max_size=999999,
                max_batch_size=99999,
                max_batch_time=0.01,
                prefetch_depth=10,
                background_flush=False,
                parallelism=1):
        self.queue_config = QueueConfig(
            max_size=max_size,
            max_batch_size=max_batch_size,
            max_batch_time=max_batch_time,
            prefetch_depth=prefetch_depth,
            background_flush=background_flush)
        self.parallelism = parallelism


class DefaultStreamingSource(object):
    """default streaming source implement"""
    def get_next(self):
        # get current record
        return None




# -*- coding: utf-8 -*-
# 1. Warm up for 70 days
# 2. Generate 5 random momentum strategies
# 3. Run these strategies for 30 days
# 4. Pick the best of the strategies that have been run, and generate 5 new strategies derived from the top performers
# 5. Continue with this process, and make trades based off the strategy that performed the best in the past
# The random strategies have 4 "genes" in their "genome"
# derivatives determines, generally speaking, if we want to take action when the price has just a positive/negative first derivative or both a positive/negative first and second derivative
# vwap1 is the main vwap
# vwap2 is used if we have derivatives = 2
# shortlong determines whether we go short or long on upswings/downswings
# For each strategy "in testing", search through stocks that fit the criteria and take action
# Using the strategy that performed the best, make trades
import random
from collections import deque
import zipline
from zipline.finance import commission, slippage
from zipline.api import order_target, record, order_target_percent

import pandas as pd
from collections import OrderedDict
import pytz
from datetime import datetime
from contrib_lib.quant.utils.chinese_stock_calendar import ChineseStockCalendar
from contrib_lib.quant.zipline.default_strategy import DefaultStrategy


class EvolutionaryStrategy(DefaultStrategy):
    """
    进化算法策略
    参考：https://www.quantopian.com/posts/evolutionary-strategy
    """
    def __init__(self):
        super().__init__()

    def prepare_data(self):
        data = OrderedDict()
        data['SPY'] = pd.read_csv('/Users/tony/zipline_data_can_delete/AAPL.csv', index_col=0,
                                  parse_dates=[['Date', 'Timestamp']])
        data['SPY'] = data['SPY'][['OpenPrice', 'HighPrice', 'LowPrice', 'ClosePrice', 'TotalVolume']]
        panel = pd.Panel(data)
        panel.major_axis = panel.major_axis.tz_localize(pytz.utc)
        panel.minor_axis = ['open', 'high', 'low', 'close', 'volume']
        return panel

    # Initialize the needed starting variables
    def initialize(self, context):
        context.set_slippage(slippage.FixedSlippage(spread=0))
        context.set_commission(commission.PerTrade(cost=0))
        context.target_duration = 30  # How long to test strategies for
        context.num_parents = 4  # Number of parents each new species has
        context.generation_size = 5  # How many species to generate per generation
        context.frames = deque(maxlen=70)  # Store past frames for past prices
        context.species = []  # All species are stored here
        self.new_generation(context)  # Create a new generation (starts testing on 70th day)

    # Creates a new generation of size context.generation_size based off of top performers from past generations
    def new_generation(self, context):
        possible_parents = []
        for spec in context.species:
            if spec['run_duration'] >= context.target_duration: possible_parents.append(spec)
        parents = sorted(possible_parents, key=lambda k: k['value'])[-context.num_parents:]

        for i in range(context.generation_size):
            # Create new species with 4 genes from parents and a new random gene
            random_gene = random.randrange(4)
            if random_gene == 0 or len(parents) < context.num_parents:
                derivatives = random.randint(1, 2)
            else:
                derivatives = parents[random.randrange(context.num_parents)]['derivatives']

            if random_gene == 1 or len(parents) < context.num_parents:
                vwap1 = random.randint(3, 70)
            else:
                vwap1 = parents[random.randrange(context.num_parents)]['vwap1']

            if random_gene == 2 or len(parents) < context.num_parents:
                vwap2 = random.randint(2, vwap1 - 1)
            else:
                vwap2_poss = []
                for spec in parents:
                    if spec['vwap2'] < vwap1: vwap2_poss.append(spec['vwap2'])
                if len(vwap2_poss) > 0:
                    vwap2 = random.choice(vwap2_poss)
                else:
                    vwap2 = random.randint(2, vwap1 - 1)

            if random_gene == 3 or len(parents) < context.num_parents:
                shortlong = random.randrange(2)
            else:
                shortlong = parents[random.randrange(context.num_parents)]['shortlong']

            # For each species, store its genes, how long it's been run for, its value (starts at 100, similar to $100), and its positions
            context.species.append(
                {'derivatives': derivatives, 'vwap1': vwap1, 'vwap2': vwap2, 'shortlong': shortlong, 'run_duration': 0,
                 'value': 100, 'current_longs': [], 'current_shorts': []})
        return None

    # Run a specific strategy species, used on strategies with run_duration less than context.target_duration and also the best overall returner
    def run_strategy(self, key, context, data):
        strat = context.species[key]
        for sid in data:
            if strat['derivatives'] == 1:
                # If price is more or less than vwap, take action
                if data[sid].price > self.dynamicvwap(context, sid, strat['vwap1']):
                    if strat['shortlong'] == 0 and sid not in strat['current_longs']:
                        strat['current_longs'].append(sid)
                        if sid in strat['current_shorts']:
                            strat['current_shorts'].remove(sid)
                    elif strat['shortlong'] == 1 and sid not in strat['current_shorts']:
                        strat['current_shorts'].append(sid)
                        if sid in strat['current_longs']:
                            strat['current_longs'].remove(sid)
                elif data[sid].price < self.dynamicvwap(context, sid, strat['vwap1']):
                    if strat['shortlong'] == 1 and sid not in strat['current_longs']:
                        strat['current_longs'].append(sid)
                        if sid in strat['current_shorts']:
                            strat['current_shorts'].remove(sid)
                    elif strat['shortlong'] == 0 and sid not in strat['current_shorts']:
                        strat['current_shorts'].append(sid)
                        if sid in strat['current_longs']:
                            strat['current_longs'].remove(sid)

            else:
                # If two derivatives in agreement, take action
                if data[sid].price > self.dynamicvwap(context, sid, strat['vwap2']) and self.dynamicvwap(context, sid, strat[
                    'vwap2']) > self.dynamicvwap(context, sid, strat['vwap1']):
                    if strat['shortlong'] == 0 and sid not in strat['current_longs']:
                        strat['current_longs'].append(sid)
                        if sid in strat['current_shorts']:
                            strat['current_shorts'].remove(sid)
                    elif strat['shortlong'] == 1 and sid not in strat['current_shorts']:
                        strat['current_shorts'].append(sid)
                        if sid in strat['current_longs']:
                            strat['current_longs'].remove(sid)
                elif data[sid].price < self.dynamicvwap(context, sid, strat['vwap2']) and self.dynamicvwap(context, sid, strat[
                    'vwap2']) < self.dynamicvwap(context, sid, strat['vwap1']):
                    if strat['shortlong'] == 1 and sid not in strat['current_longs']:
                        strat['current_longs'].append(sid)
                        if sid in strat['current_shorts']:
                            strat['current_shorts'].remove(sid)
                    elif strat['shortlong'] == 0 and sid not in strat['current_shorts']:
                        strat['current_shorts'].append(sid)
                        if sid in strat['current_longs']:
                            strat['current_longs'].remove(sid)
        return None

    # Updates returns after a strategy is run
    def update_returns(self, key, context, data):
        strat = context.species[key]
        daily_total_returns = 0
        for sid in data:
            if sid in strat['current_longs']:
                daily_total_returns += data[sid].returns()
            elif sid in strat['current_shorts']:
                daily_total_returns -= data[sid].returns()
        if len(strat['current_longs']) + len(strat['current_shorts']) > 0:
            strat['value'] *= (daily_total_returns / (len(strat['current_longs']) + len(strat['current_shorts'])) + 1)
        return None

    # We need a special volume-weighted average price function since our num_bars isn't static
    def dynamicvwap(self, context, sid, num_bars):
        addedprices = 0
        count = 0
        for i in range(1, num_bars + 1):
            bar = context.frames[-i]
            addedprices += bar[sid].price
            count += 1
        return addedprices / count

    # This is called once per bar
    def handle_data(self, context, data):
        context.frames.append(data)  # Update the recent bars
        top_returner = {}

        # If we have enough past data:
        if len(context.frames) == 70:
            new_gen = 1
            # Find the top performer
            tested_strats = []
            for strat in context.species:
                if strat['run_duration'] >= context.target_duration:
                    tested_strats.append(strat)
            top_returners = sorted(tested_strats, key=lambda k: k['value'])
            if top_returners:
                top_returner = top_returners[-1]
            # For each species, check if it needs to be run and then run it
            for i in range(len(context.species)):
                if context.species[i]['run_duration'] < context.target_duration or context.species[i] == top_returner:

                    # Update returns, but not for the top performer to prevent bias
                    if context.species[i] != top_returner:
                        self.update_returns(i, context, data)
                        context.species[i]['run_duration'] += 1
                        new_gen = 0
                    self.run_strategy(i, context, data)

            if new_gen == 1:
                self.log.info('Current top performer:')
                self.log.info('derivatives: ' + str(top_returner['derivatives']))
                self.log.info('vwap1: ' + str(top_returner['vwap1']))
                self.log.info('vwap2: ' + str(top_returner['vwap2']))
                self.log.info('shortlong: ' + str(top_returner['shortlong']))
                self.log.info('# of positions: ' + str(len(top_returner['current_shorts']) + len(top_returner['current_longs'])))
                self.log.info('Value: ' + str(top_returner['value']))
                self.log.info('-------------')
                self.new_generation(context)  # Create new generation

            longs = []
            shorts = []
            context.portfolio.portfolio_value
            # Make trades based on the top performer
            # Buy and sell shares in 2:1 leverage with amount proportional to our starting cash
            if top_returner:
                for sid in data:
                    if sid in top_returner['current_longs']:
                        longs.append(sid)
                    elif sid in top_returner['current_shorts']:
                        shorts.append(sid)

                for sid in data:
                    if sid in longs:
                        order_target_percent(sid, 1.0 / len(longs))
                    elif sid in shorts:
                        order_target_percent(sid, -1.0 / len(shorts))
                    else:
                        order_target(sid, 0)
        try:
            top_returner['value']
        except:
            top_returner['value'] = 0
        record(num_species=len(context.species), top_value=top_returner['value'])

    def run_algorithm(self):
        # start_time = datetime(2018, 1, 2, 9, 31, 0, 0, pytz.utc)
        start_time = datetime(2018, 1, 5, 9, 31, 0, 0, pytz.utc)
        end_time = datetime(2018, 2, 4, 16, 0, 0, 0, pytz.utc)
        zipline.run_algorithm(start=start_time,
                             end=end_time,
                             initialize=self.initialize,
                             capital_base=100000,
                             handle_data=self.handle_data,
                             before_trading_start=self.before_trading_start,
                             data_frequency='minute',
                             data=self.prepare_data(),
                             trading_calendar=ChineseStockCalendar(),
                             analyze=self.analyze)


if __name__ == '__main__':
    EvolutionaryStrategy().run_algorithm()




# -*- coding: utf-8 -*-
import logging
import time
import threading
from ray.experimental.streaming.streaming import DataStream
from py_common_util.common.date_utils import DateUtils
from skydl.ray.experimental.ray_ring_buffer import RayRingBuffer
from skydl.ray.experimental.ray_streaming_executor import RayStreamingExecutor
from skydl.ray.experimental.ray_streaming_util import DefaultStreamingSource

# define logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class RayStreamingExecutorSample(RayStreamingExecutor):
    """
    ray流计算执行器sample
    """
    @property
    def ray_ring_buffer(self):
        return self._ray_ring_buffer

    def __init__(self):
        super().__init__()
        self._ray_ring_buffer = RayRingBuffer(1000)

        def exec_forever():
            try:
                count = 0
                while True:
                    count += 1
                    self.ray_ring_buffer.put(str(count) + "*aaa")
                    time.sleep(4)
            except Exception as e:
                logger.error("error occurred at exec_forever ringbuffer put, %s", str(e))
        put_ring_buffer_thread = threading.Thread(target=exec_forever, args=())
        put_ring_buffer_thread.setDaemon(True)
        put_ring_buffer_thread.start()

    def build_stream(self, streaming_parallelism, env) -> DataStream:
        class RayStreamingSource(DefaultStreamingSource):
            @property
            def global_queue(self):
                return self._global_queue

            @property
            def ringbuffer(self):
                return self._ringbuffer

            def __init__(self, ray_ring_buffer):
                self._count = 0
                self._ringbuffer = ray_ring_buffer

            def get_next(self):
                self._count += 1
                print("source will get something...")
                something = self.ringbuffer.get()
                print("source got something=" + something)
                time.sleep(0.5)
                return str(self._count) + "-" + something

        def splitter_fn(record):
            """convert record to another record list, e.g. [record]"""
            logger.info("splitter: " + record)
            # return [record+"-splitter-" + str(i) for i in range(20)]
            return [record+"-splitter-" + DateUtils.now_to_str()]

        def handle_record_fn(record):
            """处理回测的逻辑"""
            logger.info("handle record=" + record)
            time.sleep(10)
            logger.info(record + ", finished:" + DateUtils.now_to_str())

        stream = env.source(
            RayStreamingSource(self.ray_ring_buffer)
        ).set_parallelism(5).flat_map(
            flatmap_fn=splitter_fn
        ).set_parallelism(1).map(
            map_fn=handle_record_fn,
            name="nocode_backtest_map"
        ).set_parallelism(5)
        return stream





# -*- coding: utf-8 -*-
from .abstract_strategy import AbstractStrategy


class DefaultStrategy(AbstractStrategy):
    """
    abstract_strategy.py的默认实现类
    """
    def __init__(self):
        super().__init__()

    def prepare_data(self):
        return super().prepare_data()

    def initialize(self, context):
        super().initialize(context)

    def before_trading_start(self, context, data):
        super().before_trading_start(context, data)

    def handle_data(self, context, data):
        super().handle_data(context, data)

    def analyze(self, context, records):
        super().analyze(context, records)

    def run_algorithm(self):
        return super().run_algorithm()

